#include <16F88.h>
#include "mg3_std.h"
#include "mg3_motor.h"
#include "mg3_usart.h"

#BYTE CCP1CON = 0x17
#BYTE RCSTA = 0x18

/* 本番用PI制御を行うプログラム */

static signed int8 count;	//１制御周期内回転角[deg]　！注意：速度算出には使えない！
union TotalCount{
	signed int32 count32;		/* 4byte単位で扱う＝増減時 */
	unsigned int8 count8[4];	/* 1byte単位で扱う＝送信時 */
};
static union TotalCount total_count;	//総回転角度[deg]
static unsigned int8 Kp, Ki;	//Kp : Pゲイン　Ki : Iゲイン
static signed int8 Rk;		//目標角速度[deg/周期]
static signed int8 speed;	//現在の角速度[deg/周期]
static int1 start;		//1→[スタート]状態、0→[ストップ]状態
static unsigned int16 time;	//Timer割込み周期指定用

/* INT割込み(RE) */
#int_ext
void ext_isr(void){
	if (input( C_CW )){
		count++;
		output_low( LED1 );		//回転方向確認用。
	}else{
		count--;
		output_high( LED1 );		//回転方向確認用。
	}
}

/* Timer割込み */
/*
 * PI制御を行う．
 * e(k) = r(k) - x(k)
 * u(k) = u(k-1) + Kp * (e(k) - e(k-1)) + Ki * e(k)
 *  k：現在の離散時刻（制御回数）
 *  r(k)：目標値
 *  x(k)：モータの状態（角度）
 *  e(k)：制御誤差
 *  u(k)：制御入力
 */
#int_timer1
void timer1_isr(void){
	static signed int16 Ek_before = 0; // e(k-1)
	static signed int16 Uk=0; // u(k)
	signed int16 Ek;        // e(k)
	signed int8 Xk;        // x(k)
	int8 i;

	//タイマー周期セット。
	//20MHz(50n sec) /4(1cycle=4clock) / 8(3bit prescaler) = 625kHz(1.6u sec)
	//Timer1は16bit。よって、
	//最長 1.6[u sec] x 0x10000[count] = 104.856[m sec]　！オーバーフローで割込みなので16bitのMAX(0xFFFF)+1。

	//周期100[m sec]...1.6[u sec] x 62.5[k count] = 100m sec
	//65536 - 62500 = 3036 = 0x0BDC　！62.5k回カウントさせたいのでMAXから引いた数をカウンタの初期値にする。

	//周期100[m sec]
	set_timer1( time );

	// 割り込みが禁止されている間に，
	// ロータリエンコーダの値を保存・クリアする．
	Xk = speed = count;
	count = 0;

#ASM
	BCF		0x0C,0					//Timer1割込みフラグクリア
#ENDASM
	enable_interrupts( GLOBAL );		//全体割込み許可

	total_count.count32 += Xk;

	if( start ){
		Ek = Rk - Xk;
		Uk += Kp * (Ek - Ek_before) + Ki * Ek;
//		Uk += Kp * (Ek - Ek_before);
//		Uk += (Ki * Ek) >> 3;

		// PWMのパルス幅が，周期を超えないように調整する．
		if( Uk > 0x3FF ){
			Uk = 0x03FF;
		}
		else if( Uk < -0x3FF ){
			Uk = -0x3FF;
		}
		//Uk *= Calibrat;		//もしくはUk += Calibrat;？

		// 回転方向が同じ場合と異なる場合で，処理を分ける．
		if( ( bit_test(Rk, 7) == bit_test(Xk, 7) ) || Xk == 0x00 ){
			// 目標速度と実際のモータの回転速度の最上位bitが等しい(正負が等しい)か
			// 実際のモータの回転速度が0である．
			// つまり，同じ方向に回転を制御しようとしているか，モータが回転していない
			// この場合，求める制御をただちに行う．
			if( Uk >= 0 ){
				// 順方向への回転を行う．
				motor_face();

				// LEDの点灯（消灯）
				output_low(LED2);
			}
			if( Uk < 0 ){
				// 逆方向への回転を行う．
				motor_reverse();

				// LEDの点灯（消灯）
				output_high(LED2);
			}

			// モータの回転速度を制御する(PWMのduty比)を設定する．
			set_pwm1_duty(abs(Uk));
		}else{
			// 実際の回転方向と，制御しようとする回転方向が違う．
			// ドライバIC保護のため，回転をいったん止める必要がある．

			// モータへの電流供給を止める(duty比を0にする)．
			set_pwm1_duty(0x00);

			// PI制御への副作用を防止するために
			// Uk, Ekを0にしておく．
			// 理論的には，大ざっぱすぎるので，
			// 牛丸先生以外の専門家には見せないでください;-)
			// できれば，もっとましな対策をお願いいたします．
			Uk = Ek = 0;

			// モータの結線をショートさせ，ブレーキをかける．
			// もし，ドライバICが焼ける等の現象が起こった場合，
			// このコードは削除してください．
			// その場合，特にロボットの重量が大きい場合，
			// ロボットの逆回転制御へのレスポンスは非常に遅いものになります．
			motor_anchor();
		}
	}

#ASM
	BCF	0x0B,7					//全体割込み禁止
#ENDASM
//	' | | ,	下の関数だと永続的に割込み禁止。
//	disable_interrupts( GLOBAL );		//全体割込み禁止

	Ek_before = Ek;
}

int main( void ){
	int i;
	unsigned int8 req;	//リクエスト
	unsigned int8 res;	//レスポンス
	unsigned int8 Kp_temp, Ki_temp;	//[パラメータ設定]保管用。ゲインは0〜255
	signed int8 R_temp;	//[パラメータ設定]保管用。目標値は-128〜127
	signed int8 X_out;		//[パラメータ取得]退避用。速度は-128〜127
	unsigned int8 temp;	//[パラメータ取得]計算用・・・値域は不明。（ビット演算などに用いる）
	union PwmDuty{
		int16 duty16;
		int8 duty8[2];
	} duty;	//[パラメータ取得]Duty比をPC側に送信する際に使用。
	union TimeTemp{
		unsigned int16 time16;
		unsigned int8 time8[2];
	} time_temp;	//Timer割込み周期設定用。


/* ここからブートシーケンス。 */
/* 困った時はreset_cpu()を使えるように全てを初期化。 */
/* ブートシーケンス中の特徴。LEDが片方点灯、片方消灯で固定。REに触れても反応しない。 */
/* 逆にブートシーケンスを終えるとREを動かすとLEDが変化するようになる。 */
/* ブートシーケンス終了条件は通信エラー解除。 */
/* 単体テストでは成功する場合、FPGA回路とのver.適合を確認すること。 */
	disable_interrupts( GLOBAL);		//ココから割込み不可
	rs232_stop();	//USART(RS232)停止。
//アナログ入力不使用
//CCPモジュール：PWMモード
//タイマー1セットアップ（タイマー割込み用）
//タイマー2セットアップ（PWM用）
//初期Duty比：0% (Max 100% = 0x3FF)
	setup_adc( NO_ANALOGS );
	set_tris_a( 0b11100001 );	//RA4-1 : Motor回転方向
	motor_anchor();				//ブレーキモード。
	set_tris_b( 0b00010111 );	//RB7-6 : LED	RB5 : TxD	RB3 : PWM出力
	output_b( 0b10000000 );		//LED1 : 点灯	LED2 : 消灯	TxD : Break
	setup_ccp1( CCP_PWM );
	setup_timer_1( T1_INTERNAL | T1_DIV_BY_8 );	//Timer割込み用カウンタ設定。
	setup_timer_2( T2_DIV_BY_16, 0xFF, 1);	//PWM周期設定。(255+1)*4/20[MHz]*16 = 0.8192[msec] = 1.22[kHz]
	set_pwm1_duty( 0x00 );

//最初は[ストップ]状態
	start = 0;
//初期値。最初は全て0。
	total_count = 0;
	count = 0;
	Kp = 0;
	Ki = 0;
	Rk = 0x00;
//初期割込み周期
	time = 0x0BDC;

//起動最初のBreak ( 10m sec )
	delay_ms( 10 );
	rs232_restart();
//FPGA側が準備完了するまで待機。（Break受信がなくなるまで）
	for( i=0; i<10; ){
		delay_cycles(1);
		if( input(RS232C_RX) ){
			i++;
		}
	}

	enable_interrupts( INT_TIMER1);
	enable_interrupts( INT_EXT);
	enable_interrupts( GLOBAL);		//ココから割込み許可〜

//通信は割込みを行っていない暇な時間に行う。
	while( 1 ){
		if( kbhit() ){
			req = readch();
			switch( req ){
			case 0x00 :			//累積カウントクリア
				disable_interrupts( GLOBAL );	//値を破壊しない為に割込み禁止！
				total_count.count32 = 0;
				enable_interrupts( GLOBAL );	//割込み禁止を解除
				putchar( req | 0x80 );
				break;
			case 0x01 :			//モーターパラメータ設定
				R_temp = readch();				//予め受け取って保管・・・
				Kp_temp = readch();
				Ki_temp = readch();
				disable_interrupts( GLOBAL );	//パラメータを変更するので割込み禁止
				Rk = R_temp;
				Kp = Kp_temp;
				Ki = Ki_temp;
				enable_interrupts( GLOBAL );	//割込み禁止を解除
				putchar( req | 0x80 );
				break;
			case 0x02 :			//モーターステータス取得
				disable_interrupts( GLOBAL );
				X_out = speed;				//現在の速度を退避（通信は時間かかるから）
				duty.duty16 = CCP_1 & 0x00FF ;	//CCP_1にはDuty比が収められている。（2byte変数）
				duty.duty16 <<= 2;		//Duty比：上位8bit + 下位2bit = 10bitなので隙間空け
				temp = CCP1CON;		//4-5bitがDuty下位2bit。
				enable_interrupts( GLOBAL );
				duty.duty8[0] &= 0xFC;	//上位6bitをマスク。0xFC = 0b11111100
				temp >>= 4;			//下位2bitにする為4bit右シフト
				temp &= 0x03;			//下位2bitをマスク。
				duty.duty8[0] |= temp;	//アナタと合体（論理和）したい・・・
				putchar( req | 0x80 );		//順番に送信。
				putchar( X_out );				//速度(-128〜127)
				putchar( duty.duty8[0] );	//PWM(-0x3FF〜0x3FF)
				putchar( duty.duty8[1] );	//PICにとってforループは高負荷なので展開。
				for( i=0; i<4; i++ ){
					putchar( total_count.count8[i] );	//総カウント数
					delay_us(1000);
				}
				break;
			case 0x04 :			//スタート
				disable_interrupts( GLOBAL );
				start = 1;				//スタートフラグを立てる。（立ってないとPI制御をスキップ）
				setup_ccp1( CCP_PWM );
				enable_interrupts( GLOBAL );
				putchar( req | 0x80 );
				break;
			case 0x08 :			//ストップ
				disable_interrupts( GLOBAL );
				start = 0;				//PI制御をスキップさせる。（パラメータは一切変更されない）
				set_pwm1_duty( 0x000 );
				motor_anchor();			//ただし、アンカー（フルブレーキ）
				setup_ccp1( CCP_OFF );
				output_low( M_e );
				enable_interrupts( GLOBAL );
				putchar( req | 0x80 );
				break;
			case 0x20 :			//制御周期変更
				time_temp.time8[0] = readch();
				time_temp.time8[1] = readch();
				disable_interrupts( GLOBAL );
				time = time_temp.time16;
				enable_interrupts( GLOBAL );
			default :			//NCK
				putchar( req | 0x40 );		//コンナの知らないわよバカ—っ；＞＜。
			}
		}
	}
}

/*		if( kbhit() ){
			req = getc();
			res = req | 0x80;
			putc( res );
			printf(" %ld\n",total_count);
		}

//カウント値送信
if(k++ %5 == 0){
	for( i=0; i<4; i++ ){
		putchar( total_count.count8[i] );
		delay_us(1000);
	}
	if( input(LED2) ){
		output_low( LED2 );
	}else{
		output_high( LED2 );
	}
}
//	printf("%ld\n",total_count);
*/
